/*====================================+=====================================+
! File CDownload.h                    ! Copyright (C) 2016-2018 Remi PASCAL !
+-------------------------------------+-------------------------------------+
! This file is part of matelefree.                                          !
! matelefree is free software: you can redistribute it and/or modify it     !
! under the terms of the GNU General Public License as published by the     !
! Free Software Foundation, either version 3 of the License, or any later   !
! version.                                                                  !
! matelefree is distributed in the hope that it will be useful, but WITHOUT !
! ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or     !
! FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for !
! more details.                                                             !
! You should have received a copy of the GNU General Public License along   !
! with matelefree. If not, see <http://www.gnu.org/licenses/>.              !
+==========================================================================*/



/*-------------------------------------------------------------------------*/
#pragma once
/*-------------------------------------------------------------------------*/
#include <memory>
#include <wx/mstream.h>
#include <wx/uri.h>
/*-------------------------------------------------------------------------*/



/*-------------------------------------------------------------------------*/
class CDownloadEvent : public wxCommandEvent
{
   /*----------------------------------------------------------------------*/
   public :
      /*-------------------------------------------------------------------*/
      CDownloadEvent( wxEventType commandType = wxEVT_NULL, int winid = 0 )
                    : wxCommandEvent( commandType, winid )
      {  ; }

   /*----------------------------------------------------------------------*/
   private :
      /*-------------------------------------------------------------------*/
      wxFileOffset m_fo_download_size = -1 ;

   /*----------------------------------------------------------------------*/
   public :
      /*-------------------------------------------------------------------*/
      virtual wxEvent *Clone() const override
      {  return new CDownloadEvent( *this ) ; }
      /*-------------------------------------------------------------------*/
      int get_id() const { return( GetId() ) ; }
      void set_id( int i_id ) { SetId( i_id ) ; }
      int get_err_code() const { return( GetInt() ) ; }
      void set_err_code( int i_err_code ) { SetInt( i_err_code ) ; }
      wxString get_string() const { return( GetString() ) ; }
      void set_string( const wxString &s ) { SetString( s ) ; }
      wxFileOffset get_current_size() const { return( GetExtraLong() ) ; }
      void set_current_size( wxFileOffset fo_size )
      {  SetExtraLong( fo_size ) ; }
      wxFileOffset get_download_size() const
      {  return( m_fo_download_size ) ; }
      void set_download_size( wxFileOffset fo_size )
      {  m_fo_download_size = fo_size ; }

   /*----------------------------------------------------------------------*/
} ;

/*--------------------------------------------------------------------------+
! Download events :                                                         !
! - "Id" (GetId()) : download identifier (id) (e_download_id)               !
! - "int" (GetInt()) : the error code, if any                               !
! - "long" (GetExtraLong()) : current downloaded number of bytes            !
! - "String" (GetString()) : default destination if no output is set        !
+--------------------------------------------------------------------------*/
wxDECLARE_EVENT( EVT_DOWNLOAD, CDownloadEvent ) ;

/*-------------------------------------------------------------------------*/
class CThreadDownload ;

/*-------------------------------------------------------------------------*/
class CDownloadParam
{
   /*----------------------------------------------------------------------*/
   public :
      /*-------------------------------------------------------------------*/
      CDownloadParam( wxEvtHandler *handler,
                      const wxString &s_url, const wxFileName &fn_local,
                      wxString *p_s_data = nullptr,
                      wxMemoryOutputStream *p_mos = nullptr
                    )
                    : m_handler( handler ),
                      m_s_url( s_url ), m_fn_local( fn_local ),
                      m_p_s_data( p_s_data ),
                      m_p_mos( p_mos )
      {  ; }
      ~CDownloadParam() ;
      /*-------------------------------------------------------------------*/
      enum e_err_code
      {  DE_NO_ERROR_DONE, DE_NO_ERROR_INFO, DE_INIT_ERROR, DE_BAD_REQUEST,
         DE_INVALID_URL, DE_STOP_REQUEST, DE_CONNECT,
         DE_HOST_NOT_FOUND, DE_FILE_NOT_FOUND, DE_FORBIDDEN,
         DE_FILE_CREATION, DE_READ_ERROR, DE_WRITE_ERROR, DE_OTHER,
         DE_NB
      } ;
      /*-------------------------------------------------------------------*/
      static constexpr const char *m_st_tb_err_msg[ DE_NB ]
      = { wxTRANSLATE( "Download done" ),               // DE_NO_ERROR_DONE
          wxTRANSLATE( "Information" ),                 // DE_NO_ERROR_INFO
          wxTRANSLATE( "Initialization error" ),        // DE_INIT_ERROR
          wxTRANSLATE( "Bad request" ),                 // DE_BAD_REQUEST
          wxTRANSLATE( "Invalid URL" ),                 // DE_INVALID_URL
          wxTRANSLATE( "Download cancelled" ),          // DE_STOP_REQUEST
          wxTRANSLATE( "Connection error" ),            // DE_CONNECT
          wxTRANSLATE( "Host not found" ),              // DE_HOST_NOT_FOUND
          wxTRANSLATE( "File not found" ),              // DE_FILE_NOT_FOUND
          wxTRANSLATE( "Access forbidden" ),            // DE_FORBIDDEN
          wxTRANSLATE( "Local file creation error" ),   // DE_FILE_CREATION
          wxTRANSLATE( "Error reading remote file" ),   // DE_READ_ERROR
          wxTRANSLATE( "Error writing to local file" ), // DE_WRITE_ERROR
          wxTRANSLATE( "Unknown error" )                // DE_OTHER
        } ;
      /*-------------------------------------------------------------------*/
      using t_up_is = std::unique_ptr< wxInputStream > ;

   /*----------------------------------------------------------------------*/
   private :
      /*-------------------------------------------------------------------*/
      e_err_code     m_err_code = DE_NO_ERROR_INFO ;
      wxThread       *m_p_thread = nullptr ;
      /*--( Output stream )------------------------------------------------*/
      wxOutputStream *m_os = nullptr ;
      /*-------------------------------------------------------------------*/
      int            m_i_event_id = wxID_ANY ;
      wxEvtHandler   *m_handler = nullptr ;
      wxString       m_s_url ;
      wxFileName     m_fn_local ; // Local file name
      wxString       *m_p_s_data = nullptr ; // String output if asked
      wxMemoryOutputStream *m_p_mos = nullptr ; // Memory output if asked
      bool           m_boo_send_progress = false ;
      int            m_i_connect_timeout = -1 ;
      mtf::t_lst_string m_lst_header ; // http header to be used
      wxString       m_s_post ; // String data to post
      /*--( Default download storage )-------------------------------------*/
      wxString       m_s_def_data ;
      /*--( Last URL used (maybe "reLocation"s have been done ) )----------*/
      wxString       m_s_last_url ;

   /*----------------------------------------------------------------------*/
   public :
      /*-------------------------------------------------------------------*/
      void send_event( double dltotal = 1, double dlnow = 1 ) const ;
      /*-------------------------------------------------------------------*/
      e_err_code get_err_code() const { return( m_err_code ) ; }
      static wxString st_get_err_msg( int i_err_code )
      {  return( i_err_code < 0 || i_err_code >= DE_NB
                 ? "invalid code"
                 : wxGetTranslation( m_st_tb_err_msg[ i_err_code ] )
               ) ;
      }
      static e_err_code st_curle_to_de( void *curl, int i_curle ) ;
      void set_p_thread( wxThread *p_thread ) { m_p_thread = p_thread ; }
      wxThread *get_thread() { return( m_p_thread ) ; }
      wxEvtHandler *get_handler() const { return( m_handler ) ; }
      const wxString &get_url() const { return( m_s_url ) ; }
      void set_url( const wxString &s_url ) { m_s_url = s_url ; }
      wxFileName get_fn_local() const { return( m_fn_local ) ; }
      wxString *get_p_s_data() const { return( m_p_s_data ) ; }
      wxMemoryOutputStream *get_p_mos() const { return( m_p_mos ) ; }
      bool get_event_id() const { return( m_i_event_id ) ; }
      void set_event_id( int i_event_id )
      {  m_i_event_id = i_event_id ; }
      bool get_send_progress() const { return( m_boo_send_progress ) ; }
      void set_send_progress( bool boo_send_progress )
      {  m_boo_send_progress = boo_send_progress ; }
      int get_connect_timeout() const
      {  return( m_i_connect_timeout ) ; }
      void set_connect_timeout( int i_connect_timeout )
      {  m_i_connect_timeout = i_connect_timeout ; }
      void set_header( const mtf::t_lst_string &lst_header )
      {  m_lst_header = lst_header ; }
      void set_post( const wxString &s_post )
      {  m_s_post = s_post ; }
      const wxString &get_def_data() const { return( m_s_def_data ) ; }
      const wxString &get_last_url() const { return( m_s_last_url ) ; }
      wxOutputStream *get_output_stream() { return( m_os ) ; }
      void reset_output_stream()
      {  delete m_os ; m_os = nullptr ; }
      /*-------------------------------------------------------------------*/
      wxFileName build_local_fullpath() const ;
      e_err_code download() ;

   /*----------------------------------------------------------------------*/
} ;

/*-------------------------------------------------------------------------*/
class CDownload
{
   /*----------------------------------------------------------------------*/
   public :
      /*--( Download on disk )---------------------------------------------*/
      CDownload( wxEvtHandler *handler,
                 const wxString &s_url, const wxFileName &fn_local
               )
               : m_param( handler, s_url, fn_local, nullptr, nullptr )
      {  ; }
      /*--( Download in a wxString )---------------------------------------*/
      CDownload( wxEvtHandler *handler,
                 const wxString &s_url, wxString *p_s_data
               )
               : m_param( handler, s_url, wxFileName(), p_s_data, nullptr )
      {  ; }
      /*--( Download in a wxMemoryOutputStream )---------------------------*/
      CDownload( wxEvtHandler *handler,
                 const wxString &s_url, wxMemoryOutputStream *p_mos
               )
               : m_param( handler, s_url, wxFileName(), nullptr, p_mos )
      {  ; }
      /*--( Default : download in "local" wxString : cf get_def_data )-----*/
      CDownload( wxEvtHandler *handler, const wxString &s_url )
               : m_param( handler, s_url, wxFileName(), nullptr, nullptr )
      {  ; }
      /*--( Download in a wxString )---------------------------------------*/
      CDownload( const wxString &s_url, wxString &s_data )
               : m_param( nullptr, s_url, wxFileName(), &s_data, nullptr )
      {  ; }
      /*--( Download in a wxMemoryOutputStream )---------------------------*/
      CDownload( const wxString &s_url, wxMemoryOutputStream &mos )
               : m_param( nullptr, s_url, wxFileName(), nullptr, &mos )
      {  ; }
      /*--( Simpliest : download in "local" wxString : cf get_def_data )---*/
      CDownload( const wxString &s_url )
               : m_param( nullptr, s_url, wxFileName(), nullptr, nullptr )
      {  ; }
      /*-------------------------------------------------------------------*/
      ~CDownload() ;

   /*----------------------------------------------------------------------*/
   private :
      /*-------------------------------------------------------------------*/
      CDownloadParam m_param ;
      /*-------------------------------------------------------------------*/
      CThreadDownload   *m_p_thread = nullptr ;
      wxCriticalSection m_csec_p_thread ;
      bool              m_boo_thread_stopping = false ;

   /*----------------------------------------------------------------------*/
   public :
      /*-------------------------------------------------------------------*/
      static void st_init_proxy( void *curl ) ;
      /*-------------------------------------------------------------------*/
      CDownloadParam &get_param() { return( m_param ) ; }
      wxString get_last_url() const { return( m_param.get_last_url() ) ; }
      wxString get_def_data() const { return( m_param.get_def_data() ) ; }
      CDownloadParam::e_err_code non_threaded_exec() ;
      void thread_start() ;
      void thread_stop() ;
      bool thread_running() ;

   /*----------------------------------------------------------------------*/
   friend class CThreadDownload ;
   /*----------------------------------------------------------------------*/
} ;

/*-------------------------------------------------------------------------*/
class CThreadDownload : public wxThread
{
   /*----------------------------------------------------------------------*/
   public :
      /*-------------------------------------------------------------------*/
      explicit CThreadDownload( CDownload *p_dl ) ;
      ~CThreadDownload() ;
   /*----------------------------------------------------------------------*/
   private :
      /*-------------------------------------------------------------------*/
      CDownload      *m_p_dl ; // Parent object
      CDownloadParam m_param ;
   /*----------------------------------------------------------------------*/
   public :
      /*-------------------------------------------------------------------*/
      void reset_p_dl() { m_p_dl = nullptr ; }
      /*--( Will be executed by "Run" )------------------------------------*/
      virtual wxThread::ExitCode Entry() override ;

   /*----------------------------------------------------------------------*/
} ;

/*-------------------------------------------------------------------------*/
class CDlgDownload : public CDlgProgress
{
   /*----------------------------------------------------------------------*/
   public :
      /*-------------------------------------------------------------------*/
      enum e_option { DDO_NONE, DDO_END_WAIT, DDO_NB } ;
      /*-------------------------------------------------------------------*/
      CDlgDownload( wxWindow *parent, const wxString &s_url,
                    const wxFileName &fn_local,
                    e_option ddo = DDO_NONE
                  ) ;
      ~CDlgDownload() ;

   /*----------------------------------------------------------------------*/
   private :
      /*-------------------------------------------------------------------*/
      CDownload m_dl ;
      e_option m_ddo ;
      wxString  m_s_info_prev ;

   /*----------------------------------------------------------------------*/
   private :
      /*-------------------------------------------------------------------*/
      void OnBtCancel( wxCommandEvent &event ) override ;
      void OnEvent( CDownloadEvent &event ) ;
      void msg_info( const wxString &s ) ;
      void msg_error( const wxString &s ) ;

   /*----------------------------------------------------------------------*/
   public :
      /*-------------------------------------------------------------------*/
      CDownloadParam &get_param() { return( m_dl.get_param() ) ; }
      int go() ;

   /*----------------------------------------------------------------------*/
} ;

/*-------------------------------------------------------------------------*/



/*==========================================================================+
!                        End of file CDownload.h                            !
+==========================================================================*/
